#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


import sys

from airbyte_cdk import AirbyteEntrypoint, launch
from source_google_drive import SourceGoogleDrive


def run():
    args = sys.argv[1:]
    catalog_path = AirbyteEntrypoint.extract_catalog(args)
    config_path = AirbyteEntrypoint.extract_config(args)
    state_path = AirbyteEntrypoint.extract_state(args)
    source = SourceGoogleDrive(
        SourceGoogleDrive.read_catalog(catalog_path) if catalog_path else None,
        SourceGoogleDrive.read_config(config_path) if config_path else None,
        SourceGoogleDrive.read_state(state_path) if state_path else None,
    )
    launch(source, args)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import uuid
from datetime import datetime
from enum import Enum
from typing import Any, Dict, Literal, Optional, Union

import dpath.util
from pydantic.v1 import BaseModel, Field

from airbyte_cdk import OneOfOptionConfig
from airbyte_cdk.sources.file_based.config.abstract_file_based_spec import (
    AbstractFileBasedSpec,
    DeliverRawFiles,
    DeliverRecords,
)
from airbyte_cdk.sources.file_based.config.abstract_file_based_spec import (
    DeliverPermissions as DeliverPermissionsBase,
)


class RemoteIdentityType(Enum):
    USER = "user"
    GROUP = "group"


class RemoteIdentity(BaseModel):
    id: uuid.UUID
    remote_id: str
    parent_id: str | None = None
    name: str | None = None
    description: str | None = None
    email_address: str | None = None
    member_email_addresses: list[str] | None = None
    type: RemoteIdentityType
    modified_at: datetime


class RemotePermissions(BaseModel):
    id: str
    file_path: str
    allowed_identity_remote_ids: list[str] | None = None
    denied_identity_remote_ids: list[str] | None = None
    publicly_accessible: bool = False


class DeliverPermissions(DeliverPermissionsBase):
    domain: Optional[str] = Field(title="Domain", description="The Google domain of the identities.", airbyte_hidden=False, order=1)


class OAuthCredentials(BaseModel):
    class Config(OneOfOptionConfig):
        title = "Authenticate via Google (OAuth)"
        discriminator = "auth_type"

    auth_type: Literal["Client"] = Field("Client", const=True)
    client_id: str = Field(
        title="Client ID",
        description="Client ID for the Google Drive API",
        airbyte_secret=True,
    )
    client_secret: str = Field(
        title="Client Secret",
        description="Client Secret for the Google Drive API",
        airbyte_secret=True,
    )
    refresh_token: str = Field(
        title="Refresh Token",
        description="Refresh Token for the Google Drive API",
        airbyte_secret=True,
    )


class ServiceAccountCredentials(BaseModel):
    class Config(OneOfOptionConfig):
        title = "Service Account Key Authentication"
        discriminator = "auth_type"

    auth_type: Literal["Service"] = Field("Service", const=True)
    service_account_info: str = Field(
        title="Service Account Information",
        description='The JSON key of the service account to use for authorization. Read more <a href="https://cloud.google.com/iam/docs/creating-managing-service-account-keys#creating_service_account_keys">here</a>.',
        airbyte_secret=True,
    )


class SourceGoogleDriveSpec(AbstractFileBasedSpec, BaseModel):
    class Config:
        title = "Google Drive Source Spec"

    folder_url: str = Field(
        description="URL for the folder you want to sync. Using individual streams and glob patterns, it's possible to only sync a subset of all files located in the folder.",
        examples=["https://drive.google.com/drive/folders/1Xaz0vXXXX2enKnNYU5qSt9NS70gvMyYn"],
        order=0,
        pattern="^https://drive.google.com/.+",
        pattern_descriptor="https://drive.google.com/drive/folders/MY-FOLDER-ID",
    )

    delivery_method: DeliverRecords | DeliverRawFiles | DeliverPermissions = Field(
        title="Delivery Method",
        discriminator="delivery_type",
        type="object",
        order=1,
        display_type="radio",
        group="advanced",
        default="use_records_transfer",
    )

    credentials: Union[OAuthCredentials, ServiceAccountCredentials] = Field(
        title="Authentication", description="Credentials for connecting to the Google Drive API", discriminator="auth_type", type="object"
    )

    @classmethod
    def documentation_url(cls) -> str:
        return "https://docs.airbyte.com/integrations/sources/google-drive"

    @classmethod
    def schema(cls, *args: Any, **kwargs: Any) -> Dict[str, Any]:
        """
        Generates the mapping comprised of the config fields
        """
        schema = super().schema(*args, **kwargs)

        # Remove legacy settings
        dpath.util.delete(schema, "properties/streams/items/properties/legacy_prefix")
        dpath.util.delete(schema, "properties/streams/items/properties/format/oneOf/*/properties/inference_type")

        # Hide API processing option until https://github.com/airbytehq/airbyte-platform-internal/issues/10354 is fixed
        processing_options = dpath.util.get(schema, "properties/streams/items/properties/format/oneOf/4/properties/processing/oneOf")
        dpath.util.set(schema, "properties/streams/items/properties/format/oneOf/4/properties/processing/oneOf", processing_options[:1])

        return schema



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.

from airbyte_cdk.sources.file_based.exceptions import BaseFileBasedSourceError


class ErrorFetchingMetadata(BaseFileBasedSourceError):
    pass


class ErrorDownloadingFile(BaseFileBasedSourceError):
    pass



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.

import json
import logging
import uuid
from datetime import datetime
from typing import Any, Dict, Iterator, List, Tuple

import pytz
from google.oauth2 import credentials, service_account
from googleapiclient.discovery import build

from airbyte_cdk import AirbyteTracedException, FailureType
from airbyte_cdk.sources.file_based.file_based_stream_permissions_reader import AbstractFileBasedStreamPermissionsReader
from airbyte_cdk.sources.streams.core import package_name_from_class
from airbyte_cdk.sources.utils.schema_helpers import ResourceSchemaLoader
from source_google_drive.exceptions import ErrorFetchingMetadata
from source_google_drive.spec import RemoteIdentity, RemoteIdentityType, RemotePermissions, SourceGoogleDriveSpec


DRIVE_SERVICE_SCOPES = [
    "https://www.googleapis.com/auth/admin.directory.group.readonly",
    "https://www.googleapis.com/auth/admin.directory.group.member.readonly",
    "https://www.googleapis.com/auth/admin.directory.user.readonly",
]

PUBLIC_PERMISSION_IDS = [
    "anyoneWithLink",
    "anyoneCanFind",
    "domainCanFind",
    "domainWithLink",
]


def datetime_now() -> datetime:
    return datetime.now(pytz.UTC)


class SourceGoogleDriveStreamPermissionsReader(AbstractFileBasedStreamPermissionsReader):
    def __init__(self):
        super().__init__()
        self._drive_service = None
        self._directory_service = None

    @property
    def config(self) -> SourceGoogleDriveSpec:
        return self._config

    @config.setter
    def config(self, value: SourceGoogleDriveSpec):
        assert isinstance(value, SourceGoogleDriveSpec)
        self._config = value

    def _build_google_service(self, service_name: str, version: str, scopes: List[str] = None):
        if self.config is None:
            # We shouldn't hit this; config should always get set before attempting to
            # list or read files.
            raise ValueError(f"Source config is missing; cannot create the Google {service_name} client.")
        try:
            if self.config.credentials.auth_type == "Client":
                creds = credentials.Credentials.from_authorized_user_info(self.config.credentials.dict())
            else:
                creds = service_account.Credentials.from_service_account_info(
                    json.loads(self.config.credentials.service_account_info), scopes=scopes
                )
            google_service = build(service_name, version, credentials=creds)
        except Exception as e:
            raise AirbyteTracedException(
                internal_message=str(e),
                message=f"Could not authenticate with Google {service_name}. Please check your credentials.",
                failure_type=FailureType.config_error,
                exception=e,
            )

        return google_service

    @property
    def google_drive_service(self):
        if self._drive_service is None:
            self._drive_service = self._build_google_service("drive", "v3")
        return self._drive_service

    @property
    def google_directory_service(self):
        if self._directory_service is None:
            self._directory_service = self._build_google_service("admin", "directory_v1", DRIVE_SERVICE_SCOPES)
        return self._directory_service

    def _get_looping_google_api_list_response(
        self, service: Any, key: str, args: dict[str, Any], logger: logging.Logger
    ) -> Iterator[dict[str, Any]]:
        try:
            looping = True
            next_page_token: str | None = None
            while looping:
                rsp = service.list(pageToken=next_page_token, **args).execute()
                next_page_token = rsp.get("nextPageToken")
                items: list[dict[str, Any]] = rsp.get(key)

                if items is None or len(items) == 0:
                    looping = False
                    break

                if rsp.get("nextPageToken") is None:
                    looping = False
                else:
                    next_page_token = rsp.get("nextPageToken")

                for item in items:
                    yield item
        except Exception as e:
            logger.error(f"There was an error listing {key} with {args}: {str(e)}")
            raise e

    def _to_remote_file_identity(self, identity: dict[str, Any]) -> RemoteIdentity | None:
        if identity.get("id") in PUBLIC_PERMISSION_IDS:
            return None
        if identity.get("deleted") is True:
            return None

        return RemoteIdentity(
            modified_at=datetime.now(),
            id=uuid.uuid4(),
            remote_id=identity.get("emailAddress"),
            name=identity.get("name"),
            email_address=identity.get("emailAddress"),
            type=identity.get("type"),
            description=None,
        )

    def get_file_permissions(self, file_id: str, file_name: str, logger: logging.Logger) -> Tuple[List[RemoteIdentity], bool]:
        """
        Retrieves the permissions of a file in Google Drive and checks for public access.

        Args:
            file_id (str): The file to get permissions for.
            file_name (str): The name of the file to get permissions for.
            logger (logging.Logger): Logger for debugging and information.

        Returns:
            Tuple(List[RemoteFileIdentity], boolean): A list of RemoteFileIdentity objects containing permission details.
        """
        try:
            request = self.google_drive_service.permissions().list(
                fileId=file_id,
                fields="permissions, permissions/role, permissions/type, permissions/id, permissions/emailAddress",
                supportsAllDrives=True,
            )
            response = request.execute()
            permissions = response.get("permissions", [])
            is_public = False

            remote_identities = []

            for p in permissions:
                identity = self._to_remote_file_identity(p)
                if p.get("id") in PUBLIC_PERMISSION_IDS:
                    is_public = True
                if identity is not None:
                    remote_identities.append(identity)

            return remote_identities, is_public
        except Exception as e:
            raise ErrorFetchingMetadata(f"An error occurred while retrieving file permissions: {str(e)}")

    def get_file_acl_permissions(self, file: Any, logger: logging.Logger) -> Dict[str, Any]:
        remote_identities, is_public = self.get_file_permissions(file.id, file_name=file.uri, logger=logger)
        return RemotePermissions(
            id=file.id,
            file_path=file.uri,
            allowed_identity_remote_ids=[p.remote_id for p in remote_identities],
            publicly_accessible=is_public,
        ).dict(exclude_none=True)

    def load_identity_groups(self, logger: logging.Logger) -> Iterator[Dict[str, Any]]:
        domain = self.config.delivery_method.domain
        if not domain:
            logger.info("No domain provided. Trying to fetch identities from the user workspace.")
            api_args = {"customer": "my_customer"}
        else:
            api_args = {"domain": domain}

        users_api = self.google_directory_service.users()
        groups_api = self.google_directory_service.groups()
        members_api = self.google_directory_service.members()

        for user in self._get_looping_google_api_list_response(users_api, "users", args=api_args, logger=logger):
            rfp = RemoteIdentity(
                id=uuid.uuid4(),
                remote_id=user["primaryEmail"],
                name=user["name"]["fullName"] if user["name"] is not None else None,
                email_address=user["primaryEmail"],
                member_email_addresses=[x["address"] for x in user["emails"]],
                type=RemoteIdentityType.USER,
                modified_at=datetime_now(),
            )
            yield rfp.dict()

        for group in self._get_looping_google_api_list_response(groups_api, "groups", args=api_args, logger=logger):
            rfp = RemoteIdentity(
                id=uuid.uuid4(),
                remote_id=group["email"],
                name=group["name"],
                email_address=group["email"],
                type=RemoteIdentityType.GROUP,
                modified_at=datetime_now(),
            )

            for member in self._get_looping_google_api_list_response(members_api, "members", {"groupKey": group["id"]}, logger):
                rfp.member_email_addresses = rfp.member_email_addresses or []
                rfp.member_email_addresses.append(member["email"])

            yield rfp.dict()

    @property
    def file_permissions_schema(self) -> Dict[str, Any]:
        return ResourceSchemaLoader(package_name_from_class(self.__class__)).get_schema("file_permissions")

    @property
    def identities_schema(self) -> Dict[str, Any]:
        return ResourceSchemaLoader(package_name_from_class(self.__class__)).get_schema("identities")



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


from typing import Any, Mapping, Optional

from airbyte_cdk import AdvancedAuth, ConfiguredAirbyteCatalog, ConnectorSpecification, OAuthConfigSpecification, TState
from airbyte_cdk.models import AuthFlowType, OauthConnectorInputSpecification
from airbyte_cdk.sources.file_based.file_based_source import FileBasedSource
from airbyte_cdk.sources.file_based.stream.cursor.default_file_based_cursor import DefaultFileBasedCursor
from source_google_drive.spec import SourceGoogleDriveSpec
from source_google_drive.stream_permissions_reader import SourceGoogleDriveStreamPermissionsReader
from source_google_drive.stream_reader import SourceGoogleDriveStreamReader


class SourceGoogleDrive(FileBasedSource):
    def __init__(self, catalog: Optional[ConfiguredAirbyteCatalog], config: Optional[Mapping[str, Any]], state: Optional[TState]):
        super().__init__(
            stream_reader=SourceGoogleDriveStreamReader(),
            spec_class=SourceGoogleDriveSpec,
            catalog=catalog,
            config=config,
            state=state,
            cursor_cls=DefaultFileBasedCursor,
            stream_permissions_reader=SourceGoogleDriveStreamPermissionsReader(),
        )

    def spec(self, *args: Any, **kwargs: Any) -> ConnectorSpecification:
        """
        Returns the specification describing what fields can be configured by a user when setting up a file-based source.
        """
        oauth_connector_input_specification = OauthConnectorInputSpecification(
            consent_url="https://accounts.google.com/o/oauth2/v2/auth?{{client_id_param}}&{{redirect_uri_param}}&response_type=code&{{scope_param}}&access_type=offline&{{state_param}}&include_granted_scopes=true&prompt=consent",
            access_token_url="https://oauth2.googleapis.com/token?{{client_id_param}}&{{client_secret_param}}&{{auth_code_param}}&{{redirect_uri_param}}&grant_type=authorization_code",
            scope="https://www.googleapis.com/auth/drive.readonly https://www.googleapis.com/auth/admin.directory.group.readonly https://www.googleapis.com/auth/admin.directory.group.member.readonly https://www.googleapis.com/auth/admin.directory.user.readonly",
        )

        return ConnectorSpecification(
            documentationUrl=self.spec_class.documentation_url(),
            connectionSpecification=self.spec_class.schema(),
            advanced_auth=AdvancedAuth(
                auth_flow_type=AuthFlowType.oauth2_0,
                predicate_key=["credentials", "auth_type"],
                predicate_value="Client",
                oauth_config_specification=OAuthConfigSpecification(
                    oauth_connector_input_specification=oauth_connector_input_specification,
                    complete_oauth_output_specification={
                        "type": "object",
                        "additionalProperties": False,
                        "properties": {
                            "refresh_token": {
                                "type": "string",
                                "path_in_connector_config": ["credentials", "refresh_token"],
                                "path_in_oauth_response": ["refresh_token"],
                            }
                        },
                    },
                    complete_oauth_server_input_specification={
                        "type": "object",
                        "additionalProperties": False,
                        "properties": {"client_id": {"type": "string"}, "client_secret": {"type": "string"}},
                    },
                    complete_oauth_server_output_specification={
                        "type": "object",
                        "additionalProperties": False,
                        "properties": {
                            "client_id": {"type": "string", "path_in_connector_config": ["credentials", "client_id"]},
                            "client_secret": {"type": "string", "path_in_connector_config": ["credentials", "client_secret"]},
                        },
                    },
                ),
            ),
        )
